package de.fhdw.wtf.common.ast;

import de.fhdw.wtf.common.ast.type.ClassType;
import de.fhdw.wtf.common.ast.type.ProductType;
import de.fhdw.wtf.common.ast.type.Type;
import de.fhdw.wtf.common.ast.visitor.ConstructorOrOperationExceptionVisitor;
import de.fhdw.wtf.common.exception.walker.TaskException;
import de.fhdw.wtf.common.token.Token;

/**
 * This abstract class represents constructors or operations of a class.
 */
public abstract class ConstructorOrOperation extends SyntaxObject {
	
	/**
	 * Represents {@link Type} which will be returned by this constructor or operation.
	 */
	private final Type returnType;
	
	/**
	 * @return Represents the type which includes the {@link ConstructorOrOperation}
	 */
	public Type getReturnType() {
		return this.returnType;
	}
	
	private final ConstructorOrOperationSignature signature;
	
	/**
	 * Creates a new {@link ConstructorOrOperation}.
	 * 
	 * @param parameters
	 *            The parameters of the constructor or operation.
	 * @param containingType
	 *            containingType of the Constructor or Operation
	 * @param returnType
	 * @param firstToken
	 *            firstToken.
	 * @param lastToken
	 *            lastToken.
	 */
	protected ConstructorOrOperation(final String name,
			final ProductType parameters,
			final ClassType containingType,
			final Type returnType,
			final Token firstToken,
			final Token lastToken) {
		super(firstToken, lastToken);
		this.signature = ConstructorOrOperationSignature.create(name, parameters, containingType);
		this.returnType = returnType;
	}
	
	/**
	 * @return Returns the signature of this {@link ConstructorOrOperation}
	 */
	public ConstructorOrOperationSignature getSignature() {
		return this.signature;
	}
	
	@Override
	public abstract boolean equals(final Object o);
	
	@Override
	public abstract int hashCode();
	
	@Override
	public abstract String toString();
	
	/**
	 * Same as getSignature().getParameter()
	 * 
	 * @return the parameter which is assigned to this {@link ConstructorOrOperation}
	 */
	public ProductType getParameters() {
		return this.signature.getParameters();
	}
	
	/**
	 * Same as getSignature().getName()
	 * 
	 * @return the name of this {@link ConstructorOrOperation}
	 */
	public String getName() {
		return this.signature.getName();
	}
	
	/**
	 * Same as getSignature().getContainingType()
	 * 
	 * @return the containingType which is assigned to this {@link ConstructorOrOperation}
	 */
	public ClassType getContainingType() {
		return this.signature.getContainingType();
	}
	
	/**
	 * @param constructorAndOperationVisitor
	 * @throws TaskException
	 */
	public abstract void accept(final ConstructorOrOperationExceptionVisitor<TaskException> constructorAndOperationVisitor)
			throws TaskException;
}
