package de.fhdw.wtf.common.ast;

import java.util.List;

import de.fhdw.wtf.common.ast.visitor.NameReturnVisitor;
import de.fhdw.wtf.common.ast.visitor.NameVisitor;
import de.fhdw.wtf.common.ast.visitor.NameVisitorReturnBoolean;
import de.fhdw.wtf.common.ast.visitor.NameVisitorReturnName;
import de.fhdw.wtf.common.ast.visitor.NameVisitorReturnNameException;
import de.fhdw.wtf.common.exception.referencer.InvalidTypeReferenceException;
import de.fhdw.wtf.common.token.IdentifierToken;
import de.fhdw.wtf.common.token.Token;

/**
 * This class represents a {@link QualifiedName}. A {@link Name} is qualified if it contains at least two
 * {@link UnqualifiedName}s.
 */
public final class QualifiedName implements Name {
	
	/**
	 * generated.
	 */
	private static final long serialVersionUID = -8270115342124653334L;
	
	/**
	 * The rest of the name. This without first.
	 */
	private final Name rest;
	
	/**
	 * The first {@link UnqualifiedName} of the {@link QualifiedName}.
	 * 
	 */
	private final UnqualifiedName first;
	
	/**
	 * This for Multiple Inheritance.
	 */
	private final SyntaxObjectInterface thiis;
	
	/**
	 * Constructor of {@link QualifiedName}.
	 * 
	 * @param first
	 *            first part of the name
	 * @param rest
	 *            rest of the name
	 * @param identifierToken
	 *            identifierToken
	 * @param lastToken
	 *            lastToken
	 */
	private QualifiedName(final UnqualifiedName first,
			final Name rest,
			final Token identifierToken,
			final Token lastToken) {
		this.first = first;
		this.rest = rest;
		this.thiis = SyntaxObjectConcret.create(identifierToken, lastToken);
	}
	
	/**
	 * Creates a {@link QualifiedName}-Object.
	 * 
	 * @param first
	 *            first part of the name
	 * @param rest
	 *            rest of the name
	 * @return The QualifiedName-Object
	 */
	public static QualifiedName create(final UnqualifiedName first, final Name rest) {
		return create(first, rest, first.getFirstToken(), rest.getLastToken());
	}
	
	/**
	 * Creates a {@link QualifiedName}-Object.
	 * 
	 * @param first
	 *            first part of the name
	 * @param rest
	 *            rest of the name
	 * @param identifierToken
	 *            identifierToken
	 * @param lastToken
	 *            lastToken
	 * @return The QualifiedName-Object
	 */
	public static QualifiedName create(final UnqualifiedName first,
			final Name rest,
			final Token identifierToken,
			final Token lastToken) {
		return new QualifiedName(first, rest, identifierToken, lastToken);
	}
	
	/**
	 * This method returns a {@link Name} consisting of at least one {@link UnqualifiedName}. The returned {@link Name}
	 * is the name of <code>this</code> without the last part. That is only guaranteed if the method is initially called
	 * with an empty list as parameter.
	 * 
	 * @param tokens
	 *            : List of {@link IdentifierToken}.
	 * @return : Name.
	 */
	public Name getNameWithoutLast(final List<IdentifierToken> tokens) {
		tokens.add(this.getFirst().getIdentifierToken());
		return this.getRest().visit(new NameVisitorReturnName() {
			
			@Override
			public Name handleUnqualifiedName(final UnqualifiedName unqName) {
				Name name = UnqualifiedName.create(tokens.get(0));
				for (int i = 1; i < tokens.size(); i++) {
					name = name.addName(tokens.get(i));
				}
				// name.setFirstToken(tokens.get(0));
				// name.setLastToken(tokens.get(tokens.size()-1));
				return name;
			}
			
			@Override
			public Name handleQualifiedName(final QualifiedName qName) {
				return qName.getNameWithoutLast(tokens);
			}
		});
	}
	
	@Override
	public Name visit(final NameVisitorReturnName v) {
		return v.handleQualifiedName(this);
	}
	
	@Override
	public boolean equals(final Object o) {
		if (o instanceof QualifiedName) {
			final QualifiedName other = (QualifiedName) o;
			return this.getFirst().equals(other.getFirst()) && this.getRest().equals(other.getRest());
		}
		return false;
	}
	
	@Override
	public int hashCode() {
		return this.getFirst().hashCode() ^ this.getRest().hashCode();
	}
	
	@Override
	public String toString() {
		final StringBuilder result = new StringBuilder();
		result.append(this.getFirst())
				.append(de.fhdw.wtf.common.constants.parser.AstDescriptionConstants.GREATER_SYMBOL)
				.append(this.getRest());
		return result.toString();
	}
	
	@Override
	public Name addName(final IdentifierToken identifierToken) {
		return QualifiedName.create(this.getFirst(), this.rest.addName(identifierToken));
	}
	
	@Override
	public Name addName(final IdentifierToken identifierToken, final Token lastToken) {
		return QualifiedName.create(this.getFirst(), this.rest.addName(identifierToken), identifierToken, lastToken);
	}
	
	@Override
	public Name concat(final Name name) {
		return name.visit(new NameVisitorReturnName() {
			
			@Override
			public Name handleUnqualifiedName(final UnqualifiedName unqName) {
				return QualifiedName.this.addName(unqName.getIdentifierToken());
			}
			
			@Override
			public Name handleQualifiedName(final QualifiedName qName) {
				Name result;
				result = QualifiedName.this.addName(qName.getFirst().getIdentifierToken());
				result = result.concat(qName.getRest());
				return result;
			}
		});
	}
	
	@Override
	public boolean equalsPartial(final Name name) {
		final Name last = this.getLastAddedName();
		return name.visit(new NameVisitorReturnBoolean() {
			
			@Override
			public boolean handleUnqualifiedName(final UnqualifiedName unqName) {
				return last.equals(unqName);
			}
			
			@Override
			public boolean handleQualifiedName(final QualifiedName qName) {
				return last.equals(qName.getFirst());
			}
		});
	}
	
	@Override
	public UnqualifiedName getLastAddedName() {
		return (UnqualifiedName) this.getRest().visit(new NameVisitorReturnName() {
			@Override
			public Name handleUnqualifiedName(final UnqualifiedName unqName) {
				return unqName.getLastAddedName();
			}
			
			@Override
			public Name handleQualifiedName(final QualifiedName qName) {
				return qName.getLastAddedName();
			}
		});
	}
	
	@Override
	public void visit(final NameVisitor v) throws InvalidTypeReferenceException {
		v.handle(this);
	}
	
	@Override
	public boolean visit(final NameVisitorReturnBoolean v) {
		return v.handleQualifiedName(this);
	}
	
	@Override
	public Name visit(final NameVisitorReturnNameException<InvalidTypeReferenceException> v)
			throws InvalidTypeReferenceException {
		return v.handleQualifiedName(this);
	}
	
	// // PROJECTIONS /////
	
	/**
	 * The rest of the name. This without first.
	 * 
	 * @return rest
	 */
	public Name getRest() {
		return this.rest;
	}
	
	/**
	 * The first {@link UnqualifiedName} of the {@link QualifiedName}.
	 * 
	 * @return first {@link UnqualifiedName}
	 */
	public UnqualifiedName getFirst() {
		return this.first;
	}
	
	@Override
	public Token getFirstToken() {
		return this.getThis().getFirstToken();
	}
	
	@Override
	public void setFirstToken(final Token firstToken) {
		this.getThis().setFirstToken(firstToken);
	}
	
	@Override
	public Token getLastToken() {
		return this.getThis().getLastToken();
	}
	
	@Override
	public void setLastToken(final Token lastToken) {
		this.getThis().setLastToken(lastToken);
	}
	
	/**
	 * Returned the This of the Multiple Inheritance.
	 * 
	 * @return this
	 */
	private SyntaxObjectInterface getThis() {
		return this.thiis;
	}
	
	@Override
	public <X> X visit(final NameReturnVisitor<X> v) {
		return v.handle(this);
	}
}
