package de.fhdw.wtf.dsl.scanner.modelScanner.state;

import de.fhdw.wtf.common.stream.TokenStream;
import de.fhdw.wtf.common.token.Position;
import de.fhdw.wtf.common.token.symbols.ArrowToken;
import de.fhdw.wtf.common.token.symbols.AsteriskToken;
import de.fhdw.wtf.common.token.symbols.BracketCloseToken;
import de.fhdw.wtf.common.token.symbols.BracketOpenToken;
import de.fhdw.wtf.common.token.symbols.ColonToken;
import de.fhdw.wtf.common.token.symbols.CommaToken;
import de.fhdw.wtf.common.token.symbols.CurlyBracketCloseToken;
import de.fhdw.wtf.common.token.symbols.CurlyBracketOpenToken;
import de.fhdw.wtf.common.token.symbols.DoubleSquareBracketCloseToken;
import de.fhdw.wtf.common.token.symbols.DoubleSquareBracketOpenToken;
import de.fhdw.wtf.common.token.symbols.EqualToken;
import de.fhdw.wtf.common.token.symbols.ExclamationToken;
import de.fhdw.wtf.common.token.symbols.GreaterSymbolToken;
import de.fhdw.wtf.common.token.symbols.HyphenToken;
import de.fhdw.wtf.common.token.symbols.PipeToken;
import de.fhdw.wtf.common.token.symbols.PlusSymbolToken;
import de.fhdw.wtf.common.token.symbols.SemicolonToken;
import de.fhdw.wtf.common.token.symbols.SquareBracketCloseToken;
import de.fhdw.wtf.common.token.symbols.SquareBracketOpenToken;
import de.fhdw.wtf.common.token.symbols.SymbolConstants;

/**
 * Default state of the Scanner. Recognizes single-char elements(=Symbols) by itself and (otherwise) delegates
 * processing to appropriate states.
 */
public class SymbolState extends MultiCharState {
	
	/**
	 * <b>PRIVATE</b> Constructor.
	 * 
	 * @param currentInput
	 *            initial symbol
	 * @param firstPosition
	 *            of the first char.
	 */
	protected SymbolState(final String currentInput, final Position firstPosition) {
		super(currentInput, firstPosition);
	}
	
	/**
	 * Factory for {@link SymbolState}.
	 * 
	 * @param firstPosition
	 *            first Position of the first char
	 * @return {@link SymbolState}
	 */
	public static SymbolState create(final Position firstPosition) {
		return new SymbolState("", firstPosition);
	}
	
	@Override
	public void finish(final TokenStream outputStream) {
		final int inputLength = this.getCurrentInput().length();
		final StringBuilder symbols = new StringBuilder(this.getCurrentInput());
		while (symbols.length() > 0) {
			
			// calculate start position
			final Position position =
					this.calculateNewPosition(this.getFirstPosition(), inputLength - symbols.length());
			
			final String sign = Character.toString(symbols.charAt(0));
			symbols.deleteCharAt(0);
			if (sign.equals(SymbolConstants.EQUALS)) {
				outputStream.add(EqualToken.create(position));
			} else if (sign.equals(SymbolConstants.CURLYBRACKETCLOSE)) {
				outputStream.add(CurlyBracketCloseToken.create(position));
			} else if (sign.equals(SymbolConstants.CURLYBRACKETOPEN)) {
				outputStream.add(CurlyBracketOpenToken.create(position));
			} else if (sign.equals(SymbolConstants.ASTERISK)) {
				outputStream.add(AsteriskToken.create(position));
			} else if (sign.equals(SymbolConstants.SEMICOLON)) {
				outputStream.add(SemicolonToken.create(position));
			} else if (sign.equals(SymbolConstants.GREATERSYMBOL)) {
				outputStream.add(GreaterSymbolToken.create(position));
			} else if (sign.equals(SymbolConstants.PLUS)) {
				outputStream.add(PlusSymbolToken.create(position));
			} else if (sign.equals(SymbolConstants.COLON)) {
				outputStream.add(ColonToken.create(position));
			} else if (sign.equals(SymbolConstants.PIPE)) {
				outputStream.add(PipeToken.create(position));
			} else if (sign.equals(SymbolConstants.BRACKETOPEN)) {
				outputStream.add(BracketOpenToken.create(position));
			} else if (sign.equals(SymbolConstants.BRACKETCLOSE)) {
				outputStream.add(BracketCloseToken.create(position));
			} else if (sign.equals(SymbolConstants.COMMA)) {
				outputStream.add(CommaToken.create(position));
			} else if (sign.equals(SymbolConstants.EXCLAMATION)) {
				outputStream.add(ExclamationToken.create(position));
			} else if (sign.equals(SymbolConstants.SQUAREBRACKETOPEN)) {
				// Check for SquareBracketOpen on next position
				if (symbols.length() > 0
						&& Character.toString(symbols.charAt(0)).equals(SymbolConstants.SQUAREBRACKETOPEN)) {
					outputStream.add(DoubleSquareBracketOpenToken.create(position));
					symbols.deleteCharAt(0);
				} else {
					outputStream.add(SquareBracketOpenToken.create(position));
				}
			} else if (sign.equals(SymbolConstants.SQUAREBRACKETCLOSE)) {
				// Check for SquareBracketClose on next position
				if (symbols.length() > 0
						&& Character.toString(symbols.charAt(0)).equals(SymbolConstants.SQUAREBRACKETCLOSE)) {
					outputStream.add(DoubleSquareBracketCloseToken.create(position));
					symbols.deleteCharAt(0);
				} else {
					outputStream.add(SquareBracketCloseToken.create(position));
				}
			} else if (sign.equals(SymbolConstants.HYPHEN)) {
				if (symbols.length() > 0 && Character.toString(symbols.charAt(0)).equals(SymbolConstants.GREATERSYMBOL)) {
					outputStream.add(ArrowToken.create(position));
					symbols.deleteCharAt(0);
				} else {
					outputStream.add(HyphenToken.create(position));
				}
			}
		}
	}
	
	private Position calculateNewPosition(final Position oldPosition, final int count) {
		final int lineNumber = oldPosition.getLineNumber();
		final int columnNumber = oldPosition.getColumnNumber() + count;
		final int p = oldPosition.getPosition() + count;
		final String filePath = oldPosition.getFilePath();
		return Position.create(filePath, lineNumber, columnNumber, p);
	}
	
	/**
	 * Tests <code>input</code> as symbol of <b>WTF-DSL</b>.
	 * 
	 * @param input
	 *            Character to test
	 * @return <code>true</code> if <code>input</code> is a symbol of the <b>WTF-DSL</b> <br>
	 *         <code>false</code> if <code>input</code> is <b>not</b> a symbol of the <b>WTF-DSL</b>
	 */
	public static boolean isSymbol(final char input) {
		final String inputValue = String.valueOf(input);
		return inputValue.equals(SymbolConstants.EQUALS) || inputValue.equals(SymbolConstants.CURLYBRACKETCLOSE)
				|| inputValue.equals(SymbolConstants.CURLYBRACKETOPEN) || inputValue.equals(SymbolConstants.ASTERISK)
				|| inputValue.equals(SymbolConstants.SEMICOLON)
				|| inputValue.equals(SymbolConstants.SQUAREBRACKETCLOSE)
				|| inputValue.equals(SymbolConstants.SQUAREBRACKETOPEN)
				|| inputValue.equals(SymbolConstants.GREATERSYMBOL) || inputValue.equals(SymbolConstants.PLUS)
				|| inputValue.equals(SymbolConstants.COLON) || inputValue.equals(SymbolConstants.HYPHEN)
				|| inputValue.equals(SymbolConstants.PIPE) || inputValue.equals(SymbolConstants.BRACKETOPEN)
				|| inputValue.equals(SymbolConstants.BRACKETCLOSE) || inputValue.equals(SymbolConstants.COMMA)
				|| inputValue.equals(SymbolConstants.EXCLAMATION);
	}
	
	@Override
	public boolean accept(final ScannerStateVisitor visitor) {
		return visitor.handleSymbolState();
	}
}
