package de.fhdw.wtf.parser;

import de.fhdw.wtf.common.exception.parser.NoArrowException;
import de.fhdw.wtf.common.exception.parser.NoBracketCloseException;
import de.fhdw.wtf.common.exception.parser.NoBracketOpenException;
import de.fhdw.wtf.common.exception.parser.NoColonException;
import de.fhdw.wtf.common.exception.parser.NoCurlyBracketCloseException;
import de.fhdw.wtf.common.exception.parser.NoCurlyBracketOpenException;
import de.fhdw.wtf.common.exception.parser.NoDoubleSquareBracketCloseException;
import de.fhdw.wtf.common.exception.parser.NoEqualException;
import de.fhdw.wtf.common.exception.parser.NoGroupElementException;
import de.fhdw.wtf.common.exception.parser.NoGroupException;
import de.fhdw.wtf.common.exception.parser.NoIdentifierException;
import de.fhdw.wtf.common.exception.parser.NoPlusSymbolException;
import de.fhdw.wtf.common.exception.parser.NoSemicolonException;
import de.fhdw.wtf.common.exception.parser.NoSquareBracketCloseException;
import de.fhdw.wtf.common.exception.parser.NoSquareBracketOpenException;
import de.fhdw.wtf.common.stream.TokenStream;
import de.fhdw.wtf.common.token.IdentifierToken;
import de.fhdw.wtf.common.token.Token;
import de.fhdw.wtf.common.token.symbols.PlusSymbolToken;

/**
 * Contains util methods for Parser.
 */
final class ParserUtils {
	
	/**
	 * Constructor of {@link ParserUtils}.
	 */
	private ParserUtils() {
	}
	
	/**
	 * Takes the next Token from this.stream and checks if the Token is an IdentifierToken. If it is, it will be removed
	 * from this.stream and returned, otherwise a NoIdentifierException will be returned.
	 * 
	 * @param stream
	 *            stream
	 * @return {@link IdentifierToken}
	 * @throws NoIdentifierException
	 *             possible Exception
	 */
	static IdentifierToken requireAndRemoveIdentifier(final TokenStream stream) throws NoIdentifierException {
		final Token token = stream.peek();
		if (token.isIdentifierToken()) {
			final IdentifierToken iToken = (IdentifierToken) token;
			stream.removeFirst();
			return iToken;
		} else {
			throw NoIdentifierException.create(token);
		}
	}
	
	/**
	 * Throws a {@link NoArrowException} if the next Token of <code>this.stream</code> is not an ArrowToken.
	 * 
	 * @param stream
	 *            stream
	 * @throws NoArrowException
	 *             possible Exception
	 */
	static void requireAndRemoveArrowToken(final TokenStream stream) throws NoArrowException {
		if (!stream.peek().isArrowToken()) {
			throw NoArrowException.create(stream.peek());
		}
		stream.next();
	}
	
	/**
	 * Throws a NoColonException if the next Token of this.stream is a not a ColonToken.
	 * 
	 * @param stream
	 *            stream
	 * @throws NoColonException
	 *             possible Exception
	 */
	static void requireAndRemoveColonToken(final TokenStream stream) throws NoColonException {
		final Token nextToken = stream.peek();
		if (!(nextToken.isColonToken())) {
			throw NoColonException.create(nextToken);
		}
		stream.removeFirst();
	}
	
	/**
	 * Throws a {@link NoBracketOpenException} if the next Token of <code>this.getStream()</code> is not a
	 * BracketOpenToken.
	 * 
	 * @param stream
	 *            stream
	 * @throws NoBracketOpenException
	 *             possible Exception
	 */
	static void requireAndRemoveBracketOpenToken(final TokenStream stream) throws NoBracketOpenException {
		final Token token = stream.peek();
		if (!token.isBracketOpenToken()) {
			throw NoBracketOpenException.create(token);
		}
		stream.removeFirst();
	}
	
	/**
	 * Throws a {@link NoBracketCloseException} if the next Token of <code>this.stream</code> is not a
	 * BracketCloseToken.
	 * 
	 * @param stream
	 *            stream
	 * @return {@link Token}
	 * @throws NoBracketCloseException
	 *             possible Exception
	 */
	static Token requireAndRemoveBracketCloseToken(final TokenStream stream) throws NoBracketCloseException {
		final Token token = stream.peek();
		if (!token.isBracketCloseToken()) {
			throw NoBracketCloseException.create(token);
		}
		return stream.next();
	}
	
	/**
	 * Throws a {@link NoSemicolonException} if the next Token of this.stream is a not a SemicolonToken.
	 * 
	 * @param stream
	 *            stream
	 * @throws NoSemicolonException
	 *             possible Exception
	 */
	static void requireSemicolonToken(final TokenStream stream) throws NoSemicolonException {
		final Token nextToken = stream.peek();
		if (!(nextToken.isSemicolonToken())) {
			throw NoSemicolonException.create(nextToken);
		}
	}
	
	/**
	 * If the next element of this.stream is not a ColonToken, the element will be removed and skipToSemicolonToken will
	 * be called again. Otherwise only the SemicolonToken will be removed.
	 * 
	 * @param stream
	 *            stream
	 */
	static void skipToSemicolonToken(final TokenStream stream) {
		// First semicolon is the reason for exception.
		// if (stream.hasNext() && stream.peek().isSemicolonToken()) {
		// stream.removeFirst();
		// }
		if (stream.hasNext() && !(stream.peek().isSemicolonToken()) && !(stream.peek().isEndToken())) {
			stream.removeFirst();
			ParserUtils.skipToSemicolonToken(stream);
		}
		if (stream.hasNext() && stream.peek().isSemicolonToken()) {
			stream.removeFirst();
		}
	}
	
	/**
	 * Throws a NoSquareBracketCloseException if the next Token of this.stream is a not a SquareBracketCloseToken.
	 * 
	 * @param stream
	 *            stream
	 * @throws NoSquareBracketCloseException
	 *             possible Exception
	 */
	static void requireAndRemoveSquareBracketCloseToken(final TokenStream stream) throws NoSquareBracketCloseException {
		final Token nextToken = stream.peek();
		if (!nextToken.isSquareBracketCloseToken()) {
			throw NoSquareBracketCloseException.create(nextToken);
		}
		stream.removeFirst();
	}
	
	/**
	 * Throws a {@link NoDoubleSquareBracketCloseException} if the next Token of this.stream is a not a
	 * DoubleSquareBracketCloseToken.
	 * 
	 * @param stream
	 *            stream
	 * @throws NoDoubleSquareBracketCloseException
	 *             possible Exception
	 */
	static void requireAndRemoveDoubleSquareBracketCloseToken(final TokenStream stream)
			throws NoDoubleSquareBracketCloseException {
		final Token nextToken = stream.peek();
		if (!nextToken.isDoubleSquareBracketCloseToken()) {
			throw NoDoubleSquareBracketCloseException.create(nextToken);
		}
		stream.removeFirst();
	}
	
	/**
	 * Throws a NoEqualException if the next Token of this.stream is not an EqualToken.
	 * 
	 * @param stream
	 *            stream
	 * @throws NoEqualException
	 *             possible Exception
	 */
	static void requireAndRemoveEqualToken(final TokenStream stream) throws NoEqualException {
		final Token nextToken = stream.peek();
		if (!(nextToken.isEqualToken())) {
			throw NoEqualException.create(nextToken);
		}
		stream.removeFirst();
	}
	
	/**
	 * Throws a {@link NoPlusSymbolException} if the next Token of this.stream is a not a {@link PlusSymbolToken}.
	 * 
	 * @param stream
	 *            stream
	 * @return {@link PlusSymbolToken}
	 * @throws NoPlusSymbolException
	 *             possible Exception
	 */
	static PlusSymbolToken requireAndRemovePlusSymbol(final TokenStream stream) throws NoPlusSymbolException {
		final Token nextToken = stream.peek();
		if (nextToken.isPlusSymbolToken()) {
			stream.removeFirst();
			return (PlusSymbolToken) nextToken;
		} else {
			throw NoPlusSymbolException.create(nextToken);
		}
	}
	
	/**
	 * Throws a NoCurlyBracketOpenException if the next Token of this.stream is not a CurlyBracketOpenToken.
	 * 
	 * @param stream
	 *            stream
	 * @throws NoCurlyBracketOpenException
	 *             possible Exception
	 */
	static void requireAndRemoveCurlyBracketOpenToken(final TokenStream stream) throws NoCurlyBracketOpenException {
		final Token nextToken = stream.peek();
		if (!(nextToken.isCurlyBracketOpenToken())) {
			throw NoCurlyBracketOpenException.create(nextToken);
		}
		stream.removeFirst();
	}
	
	/**
	 * Throws a NoCurlyBracketCloseException if the next Token of this.stream is not a CurlyBracketCloseToken.
	 * 
	 * @param stream
	 *            stream
	 * @throws NoCurlyBracketCloseException
	 *             possible Exception
	 */
	static void requireAndRemoveCurlyBracketCloseToken(final TokenStream stream) throws NoCurlyBracketCloseException {
		final Token nextToken = stream.peek();
		if (!(nextToken.isCurlyBracketCloseToken())) {
			throw NoCurlyBracketCloseException.create(nextToken);
		}
		stream.removeFirst();
	}
	
	/**
	 * Throws a NoSquareBracketOpenException if the next Token of this.stream is not a SquareBracketOpenToken.
	 * 
	 * @param stream
	 *            stream
	 * @throws NoSquareBracketOpenException
	 *             possible Exception
	 */
	static void requireAndRemoveSquareBracketOpenToken(final TokenStream stream) throws NoSquareBracketOpenException {
		final Token nextToken = stream.peek();
		if (!(nextToken.isSquareBracketOpenToken())) {
			throw NoSquareBracketOpenException.create(nextToken);
		}
		stream.removeFirst();
	}
	
	/**
	 * Throws a NoGroupElementException if the next Token of this.stream is not a GroupToken or a ClassToken.
	 * 
	 * @param stream
	 *            stream
	 * @throws NoGroupElementException
	 *             possible Exception
	 */
	static void requireGroupElement(final TokenStream stream) throws NoGroupElementException {
		final Token nextToken = stream.peek();
		if (!(nextToken.isGroupToken() || nextToken.isClassToken() || nextToken.isExceptionToken())) {
			throw NoGroupElementException.create(nextToken);
		}
	}
	
	/**
	 * Throws a NoGroupException if the next Token of this.stream is not a GroupToken.
	 * 
	 * @param stream
	 *            stream
	 * @throws NoGroupException
	 *             possible Exception
	 */
	static void requireAndRemoveGroupToken(final TokenStream stream) throws NoGroupException {
		final Token token = stream.next();
		if (!(token.isGroupToken())) {
			throw NoGroupException.create(token);
		}
	}
}
