package de.fhdw.wtf.parser;

import java.util.Collection;

import de.fhdw.wtf.common.ast.type.SumType;
import de.fhdw.wtf.common.ast.type.ThrownType;
import de.fhdw.wtf.common.ast.type.Type;
import de.fhdw.wtf.common.exception.parser.AbstractParserException;
import de.fhdw.wtf.common.exception.parser.NoCurlyBracketCloseException;
import de.fhdw.wtf.common.exception.parser.NoTypeException;
import de.fhdw.wtf.common.stream.TokenStream;
import de.fhdw.wtf.common.token.Token;

/**
 * Parser to parse the given TokenStream to {@link SumType}s.
 */
final class SumParser {
	
	/**
	 * TokenStream.
	 */
	private final TokenStream stream;
	
	/**
	 * Collection of exceptions.
	 */
	private final Collection<AbstractParserException> exceptions;
	
	/**
	 * Constructor of {@link SumParser}.
	 * 
	 * @param stream
	 *            tokenStream
	 * @param exceptions
	 *            collection of exceptions
	 */
	private SumParser(final TokenStream stream, final Collection<AbstractParserException> exceptions) {
		this.stream = stream;
		this.exceptions = exceptions;
	}
	
	/**
	 * FactoryMethod for {@link SumParser}.
	 * 
	 * @param stream
	 *            tokenStream
	 * @param exceptions
	 *            collection of exceptions
	 * @return The {@link SumParser}-Object.
	 */
	static SumParser create(final TokenStream stream, final Collection<AbstractParserException> exceptions) {
		return new SumParser(stream, exceptions);
	}
	
	/**
	 * Parses a series of tokens and creates a {@link SumType} from it. If the series of tokens does not match the
	 * grammar an Exception will be thrown.
	 * 
	 * @return SumType
	 * @throws AbstractParserException
	 *             AbstractParserException
	 */
	Type parse() throws AbstractParserException {
		final Token firstToken = this.stream.next();
		final SumType sum = SumType.create(firstToken);
		boolean nextTokenType = true;
		if (this.stream.peek().isCurlyBracketCloseToken()) {
			nextTokenType = false;
		}
		while (!this.stream.peek().isCurlyBracketCloseToken() && this.stream.hasNext() && nextTokenType) {
			
			sum.add(TypeParser.create(this.stream, this.exceptions).parse());
			
			if (this.stream.peek().isCommaToken()) {
				this.stream.removeFirst();
				nextTokenType = true;
			} else {
				nextTokenType = false;
			}
			
		}
		if (nextTokenType) {
			throw NoTypeException.create(this.stream.peek());
		}
		if (this.stream.peek().isCurlyBracketCloseToken()) {
			this.stream.next();
		} else {
			throw NoCurlyBracketCloseException.create(this.stream.peek());
		}
		sum.setLastToken(this.stream.peek());
		if (this.stream.peek().isExclamationToken()) {
			final ThrownType thrownType = ThrownType.create(sum.getFirstToken(), sum, sum.getLastToken());
			this.stream.next();
			
			return thrownType;
		}
		return sum;
	}
}
