package de.fhdw.wtf.parser;

import java.util.Collection;

import de.fhdw.wtf.common.ast.Name;
import de.fhdw.wtf.common.ast.type.ByNameState;
import de.fhdw.wtf.common.ast.type.ThrownType;
import de.fhdw.wtf.common.ast.type.Type;
import de.fhdw.wtf.common.ast.type.TypeProxy;
import de.fhdw.wtf.common.exception.parser.AbstractParserException;
import de.fhdw.wtf.common.exception.parser.NoTypeException;
import de.fhdw.wtf.common.stream.TokenStream;
import de.fhdw.wtf.common.token.Token;

/**
 * Parser to parse the given TokenStream to different Types.
 */
final class TypeParser {
	/**
	 * TokenStream.
	 */
	private final TokenStream stream;
	
	/**
	 * Collection of exceptions.
	 */
	private final Collection<AbstractParserException> exceptions;
	
	/**
	 * Constructor of {@link TypeParser}.
	 * 
	 * @param stream
	 *            tokenStream
	 * @param exceptions
	 *            collection of exceptions
	 */
	private TypeParser(final TokenStream stream, final Collection<AbstractParserException> exceptions) {
		this.stream = stream;
		this.exceptions = exceptions;
	}
	
	/**
	 * FactoryMethod for {@link TypeParser}.
	 * 
	 * @param stream
	 *            tokenStream
	 * @param exceptions
	 *            collection of exceptions
	 * @return The {@link TypeParser}-Object.
	 */
	static TypeParser create(final TokenStream stream, final Collection<AbstractParserException> exceptions) {
		return new TypeParser(stream, exceptions);
	}
	
	/**
	 * Parses a series of Tokens and creates an Attribute from it. If The series of Tokens does not match the grammar an
	 * Exception will be thrown.
	 * 
	 * @return {@link Type}
	 * @throws AbstractParserException
	 *             AbstractParserException
	 */
	Type parse() throws AbstractParserException {
		final Token firstToken = this.stream.peek();
		Type type = null;
		if (firstToken.isCurlyBracketOpenToken()) {
			type = SumParser.create(this.stream, this.exceptions).parse();
		} else if (firstToken.isBracketOpenToken()) {
			type = ProductParser.create(this.stream, this.exceptions).parse();
		} else if (firstToken.isIdentifierToken()) {
			final Name typeName = NameParser.createNameParser(this.stream).parse();
			final ByNameState typeProxyState = ByNameState.create(typeName);
			final Type typeProxy = TypeProxy.create(firstToken, typeProxyState, this.stream.peek());
			if (this.stream.peek().isExclamationToken()) {
				this.stream.next();
				type = ThrownType.create(firstToken, typeProxy, this.stream.peek());
			} else {
				type = typeProxy;
			}
		} else if (firstToken.isSquareBracketOpenToken()) {
			type = MapParser.create(this.stream, this.exceptions).parse();
		}
		
		if (this.stream.hasNext() && this.stream.peek().isAsteriskToken()) {
			return ListParser.create(this.stream).parse(firstToken, type);
		}
		if (type == null) {
			this.exceptions.add(NoTypeException.create(this.stream.peek()));
		}
		return type;
	}
}
