package de.fhdw.wtf.walker.tasks;

import java.util.HashSet;
import java.util.Set;

import de.fhdw.wtf.common.ast.Attribute;
import de.fhdw.wtf.common.ast.ConstructorOrOperation;
import de.fhdw.wtf.common.ast.Group;
import de.fhdw.wtf.common.ast.Model;
import de.fhdw.wtf.common.ast.Name;
import de.fhdw.wtf.common.ast.type.AtomicType;
import de.fhdw.wtf.common.ast.type.BaseType;
import de.fhdw.wtf.common.ast.type.ClassType;
import de.fhdw.wtf.common.ast.type.Type;
import de.fhdw.wtf.common.ast.visitor.AtomicTypeVisitorException;
import de.fhdw.wtf.common.exception.walker.CyclicInheritanceException;
import de.fhdw.wtf.common.exception.walker.TaskException;
import de.fhdw.wtf.common.task.TaskExecutor;
import de.fhdw.wtf.common.token.Token;
import de.fhdw.wtf.walker.walker.HelperUtils;
import de.fhdw.wtf.walker.walker.SimpleWalkerTask;

/**
 * This Task check the Inheritance-Structur. If this Checker found a cyclus, it will thrown a
 * {@link CyclicInheritanceException}.
 * 
 * This Checker must run after the {@link TypeReferencer}.
 */
public final class CyclicInheritanceCheck extends SimpleWalkerTask {
	/**
	 * Constructor of {@link CyclicInheritanceCheck}.
	 * 
	 * @param m
	 *            model
	 * @param taskmanager
	 *            taskmanager
	 */
	private CyclicInheritanceCheck(final Model m, final TaskExecutor taskmanager) {
		super(m, taskmanager);
	}
	
	/**
	 * Creates a {@link CyclicInheritanceCheck}-Object.
	 * 
	 * @param m
	 *            model
	 * @param taskmanager
	 *            taskmanager
	 * @return The {@link CyclicInheritanceCheck}-Object.
	 */
	public static CyclicInheritanceCheck create(final Model m, final TaskExecutor taskmanager) {
		return new CyclicInheritanceCheck(m, taskmanager);
	}
	
	@Override
	public void handleGroup(final Group g) throws TaskException {
		// Niothing to do here
	}
	
	@Override
	public void handleClass(final ClassType c) throws TaskException {
		final CyclicChecker cyclicChecker = new CyclicChecker();
		cyclicChecker.checkClassInheritance(c);
	}
	
	@Override
	public void handleAttribute(final Attribute a, final ClassType owner) throws TaskException {
		// Nothing to do here
	}
	
	@Override
	public void finalizeTask() throws TaskException {
		// Nothing to do here
	}
	
	/**
	 * CyclicChecker check the Inheritance-Structur. If the CyclicChecker found a cyclus, it will thrown a
	 * {@link CyclicInheritanceException}.
	 */
	private class CyclicChecker {
		
		/**
		 * Found supertypes.
		 */
		private final Set<Name> foundSupertypes;
		
		/**
		 * Intern-Constructor of {@link CyclicChecker}.
		 * 
		 * @param alreadyFoundSupertypes
		 *            alreadyFoundSupertypes in earlier steps.
		 */
		CyclicChecker(final Set<Name> alreadyFoundSupertypes) {
			this.foundSupertypes = new HashSet<>();
			this.foundSupertypes.addAll(alreadyFoundSupertypes);
		}
		
		/**
		 * Constructor of {@link CyclicChecker}.
		 */
		CyclicChecker() {
			this.foundSupertypes = new HashSet<>();
		}
		
		/**
		 * Check the Inheritance-Structur of c for cycles. First- and LastToken will need for Exception-Message. They
		 * are need to mark the Failure in the wtf-Model (f.e. ...:class=Person+{... and then the Person-Position)
		 * 
		 * @param c
		 *            class
		 * @param firstToken
		 *            firstToken
		 * @param lastToken
		 *            lastToken
		 * @throws CyclicInheritanceException
		 *             CyclicInheritanceException
		 */
		void checkClassInheritance(final ClassType c, final Token firstToken, final Token lastToken)
				throws CyclicInheritanceException {
			if (this.foundSupertypes.contains(c.getTypeName())) {
				throw CyclicInheritanceException.create(c.getTypeName(), firstToken, lastToken);
			}
			this.checkClassInheritance(c);
		}
		
		/**
		 * Check the Inheritance-Structur of c for cycles.
		 * 
		 * @param c
		 *            class
		 * @throws CyclicInheritanceException
		 *             CyclicInheritanceException
		 */
		public void checkClassInheritance(final ClassType c) throws CyclicInheritanceException {
			this.foundSupertypes.add(c.getTypeName());
			
			for (final Type current : c.getSuperTypes()) {
				final AtomicType targetType = HelperUtils.getReferencedAtomicType(current);
				
				final CyclicChecker cyclicChecker = new CyclicChecker(this.foundSupertypes);
				
				targetType.accept(new AtomicTypeVisitorException<CyclicInheritanceException>() {
					
					@Override
					public void handle(final BaseType baseType) throws CyclicInheritanceException {
						// Nothing to do, this should check an other Checker
					}
					
					@Override
					public void handle(final ClassType clazz) throws CyclicInheritanceException {
						cyclicChecker.checkClassInheritance(clazz, current.getFirstToken(), current.getLastToken());
					}
				});
			}
		}
	}
	
	@Override
	public void beginTask() throws TaskException {
		// Nothing to do here
	}
	
	@Override
	public void handleConstructorOrOperation(final ConstructorOrOperation coo, final ClassType owner)
			throws TaskException {
		// Nothing to do here
		
	}
}
