package de.fhdw.wtf.walker.tasks;

import java.util.ArrayList;
import java.util.List;

import de.fhdw.wtf.common.ast.Attribute;
import de.fhdw.wtf.common.ast.Constructor;
import de.fhdw.wtf.common.ast.ConstructorOrOperation;
import de.fhdw.wtf.common.ast.Group;
import de.fhdw.wtf.common.ast.Model;
import de.fhdw.wtf.common.ast.Operation;
import de.fhdw.wtf.common.ast.UnqualifiedName;
import de.fhdw.wtf.common.ast.type.AtomicType;
import de.fhdw.wtf.common.ast.type.ByNameState;
import de.fhdw.wtf.common.ast.type.ByReferenceState;
import de.fhdw.wtf.common.ast.type.ClassType;
import de.fhdw.wtf.common.ast.type.CompositeType;
import de.fhdw.wtf.common.ast.type.InvalidState;
import de.fhdw.wtf.common.ast.type.ListType;
import de.fhdw.wtf.common.ast.type.MapType;
import de.fhdw.wtf.common.ast.type.ProductElementType;
import de.fhdw.wtf.common.ast.type.ProductType;
import de.fhdw.wtf.common.ast.type.SumType;
import de.fhdw.wtf.common.ast.type.ThrownType;
import de.fhdw.wtf.common.ast.type.Type;
import de.fhdw.wtf.common.ast.type.TypeProxy;
import de.fhdw.wtf.common.ast.visitor.CompositeTypeVisitor;
import de.fhdw.wtf.common.ast.visitor.CompositeTypeVisitorReturn;
import de.fhdw.wtf.common.ast.visitor.ConstructorOrOperationExceptionVisitor;
import de.fhdw.wtf.common.ast.visitor.TypeProxyStateVisitor;
import de.fhdw.wtf.common.ast.visitor.TypeVisitor;
import de.fhdw.wtf.common.ast.visitor.TypeVisitorReturn;
import de.fhdw.wtf.common.constants.referencer.BaseTypeConstants;
import de.fhdw.wtf.common.exception.ast.ASTException;
import de.fhdw.wtf.common.exception.parser.SumIsAnythingException;
import de.fhdw.wtf.common.exception.walker.TaskException;
import de.fhdw.wtf.common.task.TaskExecutor;
import de.fhdw.wtf.common.token.DummyToken;
import de.fhdw.wtf.common.token.IdentifierToken;
import de.fhdw.wtf.walker.walker.HelperUtils;
import de.fhdw.wtf.walker.walker.SimpleWalkerTask;

/**
 * This task sets the Prototype for every {@link CompositeType}. A Prototype is equal to the {@link Type}, but with
 * sorted Elements. So it is a synthetic Type and has no Position-Information (First- And LastToken = End-Token).
 * Furthermore two or more Types, which are equal, reference to the same Prototype. So this Task eliminate duplicate
 * {@link SumType}s, {@link ProductType}s, {@link MapType}s and Lists.
 * 
 * This Checker must run after the {@link SubtypesFillTask}.
 */
public final class PrototypesTask extends SimpleWalkerTask {
	
	/**
	 * All created prototypes.
	 */
	private final List<Type> prototypes;
	
	/**
	 * All created abstractProducts.
	 */
	private final List<ProductType> abstractProducts;
	
	/**
	 * Constructor of {@link PrototypesTask}.
	 * 
	 * @param m
	 *            Model
	 * @param taskmanager
	 *            Taskmanager
	 */
	private PrototypesTask(final Model m, final TaskExecutor taskmanager) {
		super(m, taskmanager);
		this.prototypes = new ArrayList<>();
		this.abstractProducts = new ArrayList<>();
	}
	
	/**
	 * Creates a {@link PrototypesTask}-Object.
	 * 
	 * @param m
	 *            Model
	 * @param taskmanager
	 *            Taskmanager
	 * @return The PrototypesTask-Object
	 */
	public static PrototypesTask create(final Model m, final TaskExecutor taskmanager) {
		return new PrototypesTask(m, taskmanager);
	}
	
	@Override
	public void handleClass(final ClassType c) throws TaskException {
		this.getPrototypeFor(c);
	}
	
	@Override
	public void handleGroup(final Group g) throws TaskException {
		// Nothing to do
	}
	
	@Override
	public void handleAttribute(final Attribute a, final ClassType owner) throws TaskException {
		this.getPrototypeFor(a.getAttrType());
	}
	
	@Override
	public void handleConstructorOrOperation(final ConstructorOrOperation coo, final ClassType owner)
			throws TaskException {
		coo.accept(new ConstructorOrOperationExceptionVisitor<TaskException>() {
			
			@Override
			public void handleOperation(final Operation operation) throws TaskException {
				PrototypesTask.this.getPrototypeFor(operation.getParameters());
				PrototypesTask.this.getPrototypeFor(operation.getReturnType());
				
			}
			
			@Override
			public void handleConstructor(final Constructor constructor) throws TaskException {
				PrototypesTask.this.getPrototypeFor(constructor.getParameters());
				
			}
		});
		
	}
	
	@Override
	public void beginTask() throws TaskException {
		// Create Prototype for Anything, String and Integer
		this.getModel().getAnything().setPrototype(this.getPrototypeFor(this.deleteAnythinginAnything()));
		this.getPrototypeFor(this.getModel().getString());
		this.getPrototypeFor(this.getModel().getInteger());
	}
	
	@Override
	public void finalizeTask() throws TaskException {
		final SumType anything = this.getModel().getAnything().getPrototype().accept(new TypeVisitorReturn<SumType>() {
			private static final String anythingSum = "Anything must be a Sum";
			
			@Override
			public SumType handle(final AtomicType atomicType) {
				throw new ASTException(this.getAnythingSum());
			}
			
			@Override
			public SumType handle(final CompositeType compositeType) {
				return compositeType.accept(new CompositeTypeVisitorReturn<SumType>() {
					
					@Override
					public SumType handle(final ListType list) {
						throw new ASTException(getAnythingSum());
					}
					
					@Override
					public SumType handle(final MapType map) {
						throw new ASTException(getAnythingSum());
					}
					
					@Override
					public SumType handle(final ProductType product) {
						throw new ASTException(getAnythingSum());
					}
					
					@Override
					public SumType handle(final SumType sum) {
						return sum;
					}
					
					@Override
					public SumType handle(final ThrownType thrownType) {
						throw new ASTException(getAnythingSum());
					}
				});
			}
			
			@Override
			public SumType handle(final TypeProxy typeProxy) {
				throw new ASTException(this.getAnythingSum());
			}
			
			public String getAnythingSum() {
				return anythingSum;
			}
			
		});
		// Add abstract Products to Anything
		for (final ProductType current : this.getAbstractProducts()) {
			anything.getElements().add(current);
		}
	}
	
	/**
	 * Returned Anything without Anything in Sub-Sums.
	 * 
	 * @return Anything without Anything in Sub-Sums.
	 */
	private SumType deleteAnythinginAnything() {
		final SumType anything = this.getModel().getAnything();
		ArrayList<Type> copy = new ArrayList<>();
		copy.addAll(anything.getElements());
		boolean noSumInItAnymore = false;
		while (!noSumInItAnymore) {
			boolean foundSum = false;
			final ArrayList<Type> cleanAnything = new ArrayList<>();
			for (final Type type : copy) {
				final Boolean wasSum = type.accept(new TypeVisitorReturn<Boolean>() {
					
					@Override
					public Boolean handle(final TypeProxy s) {
						s.getState().accept(new TypeProxyStateVisitor() {
							
							@Override
							public void handle(final ByNameState byName) {
								if (!byName.getName().equals(
										UnqualifiedName.create(IdentifierToken.create(
												BaseTypeConstants.ANYTHING_CONSTANT,
												DummyToken.getDummyPosition())))) {
									cleanAnything.add(s);
								}
							}
							
							@Override
							public void handle(final InvalidState invalid) {
								return;
							}
							
							@Override
							public void handle(final ByReferenceState byType) {
								if (!byType.getTargetName().equals(
										UnqualifiedName.create(IdentifierToken.create(
												BaseTypeConstants.ANYTHING_CONSTANT,
												DummyToken.getDummyPosition())))) {
									cleanAnything.add(s);
								}
							}
						});
						return false;
					}
					
					@Override
					public Boolean handle(final CompositeType c) {
						return c.accept(new CompositeTypeVisitorReturn<Boolean>() {
							
							@Override
							public Boolean handle(final SumType sum) {
								for (final Type type2 : sum.getElements()) {
									cleanAnything.add(type2);
								}
								return true;
							}
							
							@Override
							public Boolean handle(final ProductType product) {
								cleanAnything.add(product);
								return false;
							}
							
							@Override
							public Boolean handle(final MapType map) {
								cleanAnything.add(map);
								return false;
							}
							
							@Override
							public Boolean handle(final ListType list) {
								cleanAnything.add(list);
								return false;
							}
							
							@Override
							public Boolean handle(final ThrownType thrownType) {
								cleanAnything.add(thrownType);
								return false;
							}
						});
					}
					
					@Override
					public Boolean handle(final AtomicType s) {
						throw new ASTException("Thats not allowed in Sums");
					}
				});
				
				if (wasSum) {
					foundSum = true;
				}
			}
			copy = cleanAnything;
			noSumInItAnymore = !foundSum;
		}
		
		final SumType cleanAnything = SumType.create(DummyToken.getInstance(), DummyToken.getInstance());
		for (final Type type : copy) {
			cleanAnything.add(type);
		}
		
		return cleanAnything;
	}
	
	/**
	 * Liefert zum Type den Prototypen. Setzt außerdem den Prototypen beim Type und bei dessen enthaltenen Typen
	 * ebenfalls.
	 * 
	 * @param type
	 *            type
	 * @return Prototype
	 */
	protected Type getPrototypeFor(final Type type) {
		return this.getPrototypeFor(type, false);
	}
	
	/**
	 * Liefert zum Type den Prototypen. Setzt außerdem den Prototypen beim Type und bei dessen enthaltenen Typen
	 * ebenfalls.
	 * 
	 * @param type
	 *            type
	 * @param createAnAbstractProduct
	 *            If this is true, then the Method creates for Products an abstract Prototype, that means that the
	 *            attributes have also standardized names. If it false, a normal prototype will be returned.
	 * @return Prototype
	 */
	protected Type getPrototypeFor(final Type type, final boolean createAnAbstractProduct) {
		return type.accept(new TypeVisitorReturn<Type>() {
			
			@Override
			public Type handle(final AtomicType s) {
				s.setPrototype(s);
				return s;
			}
			
			@Override
			public Type handle(final CompositeType c) {
				
				return c.accept(new CompositeTypeVisitorReturn<Type>() {
					
					@Override
					public Type handle(final ListType list) {
						return PrototypesTask.this.createListPrototype(list);
					}
					
					@Override
					public Type handle(final MapType map) {
						
						return PrototypesTask.this.createMapPrototype(map);
					}
					
					@Override
					public Type handle(final ProductType product) {
						
						return PrototypesTask.this.createProductPrototype(product, createAnAbstractProduct);
					}
					
					@Override
					public Type handle(final SumType sum) {
						
						return PrototypesTask.this.createSumPrototype(sum);
					}
					
					@Override
					public Type handle(final ThrownType thrownType) {
						
						return PrototypesTask.this.createIfNeccessaryAThrownPrototype(thrownType);
					}
				});
			}
			
			@Override
			public Type handle(final TypeProxy s) {
				s.setPrototype(s);
				return s;
			}
		});
	}
	
	/**
	 * Creates a Prototype for a List.
	 * 
	 * @param list
	 *            list
	 * @return prototype of the list
	 */
	protected ListType createListPrototype(final ListType list) {
		// Prototyp erstellen
		final ListType createdPrototype =
				ListType.create(
						DummyToken.getInstance(),
						PrototypesTask.this.getPrototypeFor(list.getOf()),
						DummyToken.getInstance());
		
		// Prototyp bei Prototyp setzen
		createdPrototype.setPrototype(createdPrototype);
		
		ListType listPrototype;
		
		final Type foundPrototype = PrototypesTask.this.findPrototype(createdPrototype);
		if (foundPrototype == null) {
			// Prototyp ist neu
			PrototypesTask.this.getPrototypes().add(createdPrototype);
			listPrototype = createdPrototype;
		} else {
			// Prototyp gibt es schon
			listPrototype = (ListType) foundPrototype;
		}
		
		// Prototyp bei List setzen
		list.setPrototype(listPrototype);
		
		return listPrototype;
	}
	
	/**
	 * Creates a Prototype for a Map.
	 * 
	 * @param map
	 *            map
	 * @return prototype of the map
	 */
	protected MapType createMapPrototype(final MapType map) {
		// Prototyp erstellen
		final MapType createdPrototype =
				MapType.create(
						DummyToken.getInstance(),
						PrototypesTask.this.getPrototypeFor(map.getKey()),
						PrototypesTask.this.getPrototypeFor(map.getOf()),
						DummyToken.getInstance());
		
		// Prototyp bei Prototyp setzen
		createdPrototype.setPrototype(createdPrototype);
		
		MapType mapPrototype;
		
		final Type foundPrototype = PrototypesTask.this.findPrototype(createdPrototype);
		if (foundPrototype == null) {
			// Prototyp ist neu
			PrototypesTask.this.getPrototypes().add(createdPrototype);
			mapPrototype = createdPrototype;
		} else {
			// Prototyp gibt es schon
			mapPrototype = (MapType) foundPrototype;
		}
		
		// Prototyp bei Map setzen
		map.setPrototype(mapPrototype);
		
		return mapPrototype;
	}
	
	/**
	 * Creates a Prototype for a Product.
	 * 
	 * @param product
	 *            product
	 * @param createAnAbstractProduct
	 *            AbstractPrototype or
	 * @return prototype of the product
	 */
	protected Type createProductPrototype(final ProductType product, final boolean createAnAbstractProduct) {
		// Prototyp erstellen
		final ProductType createdPrototype = ProductType.create(DummyToken.getInstance(), DummyToken.getInstance());
		
		// AbstractPrototyp erstellen
		final ProductType createdAbstractPrototype =
				ProductType.create(DummyToken.getInstance(), DummyToken.getInstance());
		
		final List<ProductElementType> elements = product.getElements();
		
		// Vars for Attributenames of the abstract Prototype
		final String constantForAbstractPrototypeAttributeName = "p$";
		int counter = 1;
		
		for (final ProductElementType productElement : elements) {
			// Create Prototype Elemente
			ProductElementType newProductElement =
					ProductElementType.create(
							productElement.getName(),
							PrototypesTask.this.getPrototypeFor(productElement.getType()),
							DummyToken.getInstance(),
							DummyToken.getInstance());
			createdPrototype.addElement(newProductElement);
			
			// Create AbstractPrototype Elemente
			newProductElement =
					ProductElementType.create(
							constantForAbstractPrototypeAttributeName + counter,
							PrototypesTask.this.getPrototypeFor(productElement.getType(), true),
							DummyToken.getInstance(),
							DummyToken.getInstance());
			createdAbstractPrototype.addElement(newProductElement);
			
			counter++;
		}
		
		// Prototyp bei Prototyp setzen
		createdPrototype.setPrototype(createdPrototype);
		
		// Prototype bei AbstractPrototype setzen
		createdAbstractPrototype.setPrototype(createdAbstractPrototype);
		// AbstractPrototype bei AbstractPrototype setzen
		createdAbstractPrototype.setAbstractPrototype(createdAbstractPrototype);
		
		ProductType productPrototype;
		
		Type foundPrototype = PrototypesTask.this.findPrototype(createdPrototype);
		if (foundPrototype == null) {
			// Prototyp ist neu
			PrototypesTask.this.getPrototypes().add(createdPrototype);
			productPrototype = createdPrototype;
		} else {
			// Prototyp gibt es schon
			productPrototype = (ProductType) foundPrototype;
		}
		
		// Prototyp bei Produkt setzen
		product.setPrototype(productPrototype);
		
		ProductType productAbstractPrototype;
		
		foundPrototype = PrototypesTask.this.findPrototype(createdAbstractPrototype);
		if (foundPrototype == null) {
			// AbstractPrototyp ist neu
			PrototypesTask.this.getPrototypes().add(createdAbstractPrototype);
			PrototypesTask.this.getAbstractProducts().add(createdAbstractPrototype);
			productAbstractPrototype = createdAbstractPrototype;
		} else {
			// AbstractPrototyp gibt es schon
			productAbstractPrototype = (ProductType) foundPrototype;
		}
		
		// AbstractPrototyp bei Produkt setzen
		product.setAbstractPrototype(productAbstractPrototype);
		// AsbtractPrototype bei Prototyp setzen
		productPrototype.setAbstractPrototype(productAbstractPrototype);
		
		if (createAnAbstractProduct) {
			return productAbstractPrototype;
		} else {
			return productPrototype;
		}
	}
	
	/**
	 * Creates a Prototype for a Sum.
	 * 
	 * @param sum
	 *            sum
	 * @return prototype of the sum
	 */
	protected Type createSumPrototype(final SumType sum) {
		// Prototyp erstellen
		final SumType createdSumPrototype = SumType.create(DummyToken.getInstance(), DummyToken.getInstance());
		
		List<Type> standardizeElements;
		Type sumPrototype;
		try {
			final Type standardizedType = sum.standardize();
			final SumType standardizedSum = standardizedType.accept(new TypeVisitorReturn<SumType>() {
				
				@Override
				public SumType handle(final AtomicType atomicType) {
					throw new ASTException("A standardized sum is can not an atomicType");
				}
				
				@Override
				public SumType handle(final CompositeType compositeType) {
					return compositeType.accept(new CompositeTypeVisitorReturn<SumType>() {
						
						@Override
						public SumType handle(final ListType list) {
							throw new ASTException("A standardized sum is can not an atomicType");
						}
						
						@Override
						public SumType handle(final MapType map) {
							throw new ASTException("A standardized sum is can not an atomicType");
						}
						
						@Override
						public SumType handle(final ProductType product) {
							throw new ASTException("A standardized sum is can not an atomicType");
						}
						
						@Override
						public SumType handle(final SumType sum2) {
							return sum2;
						}
						
						@Override
						public SumType handle(final ThrownType thrownType) {
							throw new ASTException("A standardized sum is can not an atomicType");
						}
					});
				}
				
				@Override
				public SumType handle(final TypeProxy typeProxy) {
					throw new ASTException("A standardized sum is can not an atomicType");
				}
			});
			
			standardizeElements = standardizedSum.getElements();
			
			for (final Type currentType : standardizeElements) {
				createdSumPrototype.add(PrototypesTask.this.getPrototypeFor(currentType));
			}
			
			// If Sum has only one Element, then this is the prototype
			final Type createdPrototype;
			if (createdSumPrototype.getElements().size() == 1) {
				createdPrototype = HelperUtils.getReferencedType(createdSumPrototype.getElements().get(0));
				// ggf. Prototyp bei Prototyp setzen
				createdPrototype.accept(new TypeVisitor() {
					
					@Override
					public void handle(final TypeProxy s) {
						throw new ASTException("Not allowed here!");
					}
					
					@Override
					public void handle(final CompositeType c) {
						c.accept(new CompositeTypeVisitor() {
							
							@Override
							public void handle(final SumType sum2) {
								throw new ASTException("That could not happened!");
							}
							
							@Override
							public void handle(final ProductType product) {
								product.setPrototype(createdPrototype);
							}
							
							@Override
							public void handle(final MapType map) {
								map.setPrototype(createdPrototype);
							}
							
							@Override
							public void handle(final ListType list) {
								list.setPrototype(createdPrototype);
							}
							
							@Override
							public void handle(final ThrownType thrownType) {
								thrownType.setPrototype(createdPrototype);
							}
						});
					}
					
					@Override
					public void handle(final AtomicType s) {
						// Nothing to do
					}
				});
			} else {
				createdPrototype = createdSumPrototype;
				// Prototyp bei Prototyp setzen
				createdSumPrototype.setPrototype(createdSumPrototype);
			}
			
			final Type foundPrototype = PrototypesTask.this.findPrototype(createdPrototype);
			if (foundPrototype == null) {
				// Prototyp ist neu
				PrototypesTask.this.getPrototypes().add(createdPrototype);
				sumPrototype = createdPrototype;
			} else {
				// Prototyp gibt es schon
				sumPrototype = foundPrototype;
			}
		} catch (final SumIsAnythingException e) {
			// Sum is equals to Anything so set Anything-Prototype as Prototype
			sumPrototype = PrototypesTask.this.getModel().getAnything().getPrototype();
		}
		
		// Prototyp bei Summe setzen
		sum.setPrototype(sumPrototype);
		
		return sumPrototype;
	}
	
	/**
	 * Creates a Prototype for a ThrownType.
	 * 
	 * @param thrownType
	 *            thrownType
	 * @return prototype of the sum
	 */
	protected Type createIfNeccessaryAThrownPrototype(final ThrownType thrownType) {
		
		Type thrownPrototype;
		
		try {
			final Type type = thrownType.standardize();
			
			thrownPrototype = type.accept(new TypeVisitorReturn<Type>() {
				
				@Override
				public Type handle(final TypeProxy typeProxy) {
					return PrototypesTask.this.getPrototypeFor(typeProxy);
				}
				
				@Override
				public Type handle(final CompositeType compositeType) {
					return compositeType.accept(new CompositeTypeVisitorReturn<Type>() {
						
						@Override
						public Type handle(final ListType list) {
							return PrototypesTask.this.getPrototypeFor(list);
						}
						
						@Override
						public Type handle(final MapType map) {
							return PrototypesTask.this.getPrototypeFor(map);
						}
						
						@Override
						public Type handle(final ProductType product) {
							return PrototypesTask.this.getPrototypeFor(product);
						}
						
						@Override
						public Type handle(final SumType sum) {
							return PrototypesTask.this.getPrototypeFor(sum);
						}
						
						@Override
						public Type handle(final ThrownType thrownType2) {
							
							final Type newPrototype;
							
							final Type underlyingPrototype =
									PrototypesTask.this.getPrototypeFor(thrownType2.getUnderlyingType());
							
							final Type createdPrototype = HelperUtils.getReferencedType(underlyingPrototype);
							
							// Prototyp erstellen
							final ThrownType createdThrownPrototype =
									ThrownType.create(
											DummyToken.getInstance(),
											createdPrototype,
											DummyToken.getInstance());
							
							final Type foundPrototype = PrototypesTask.this.findPrototype(createdThrownPrototype);
							if (foundPrototype == null) {
								// Prototyp ist neu
								PrototypesTask.this.getPrototypes().add(createdThrownPrototype);
								newPrototype = createdThrownPrototype;
							} else {
								// Prototyp gibt es schon
								newPrototype = foundPrototype;
							}
							
							newPrototype.setPrototype(newPrototype);
							
							return newPrototype;
						}
					});
				}
				
				@Override
				public Type handle(final AtomicType atomicType) {
					return PrototypesTask.this.getPrototypeFor(atomicType);
				}
			});
		} catch (final SumIsAnythingException e) {
			throw new ASTException("Anything can not be thrown.");
		}
		
		thrownType.setPrototype(thrownPrototype);
		return thrownPrototype;
	}
	
	/**
	 * Sucht in den bisherigen Prototypen nach einem passenden. Wenn keiner gefunden wird, wird null zurückgeliefert.
	 * 
	 * @param type
	 *            type
	 * @return Prototype
	 */
	protected Type findPrototype(final Type type) {
		Type foundPrototype = null;
		for (final Type prototype : this.prototypes) {
			if (prototype.equals(type)) {
				foundPrototype = prototype;
			}
		}
		return foundPrototype;
	}
	
	/**
	 * Returned the prototypes.
	 * 
	 * @return prototypes
	 */
	protected List<Type> getPrototypes() {
		return this.prototypes;
	}
	
	/**
	 * Returned the abstractProducts.
	 * 
	 * @return abstractProducts
	 */
	protected List<ProductType> getAbstractProducts() {
		return this.abstractProducts;
	}
}
